import json
import argparse
import invertedai as iai
import matplotlib.pyplot as plt

from enum import Enum
from pydantic import BaseModel
from typing import Dict, List, Optional, Union

DIAGNOSTIC_ISSUE_LIBRARY = {
    10:"Agent state index change",
    11:"Agent state modified or removed before next request",
    20:"Recurrent state index change",
    21:"Recurrent state modified or removed before next request",
    50:"Static agent properties/attributes deviation from expected range",
    51:"Static agent properties/attributes modified or removed between time steps",
    52:"Static agent properties/attributes index change",
    100:"Agents removed before next request",
    101:"Agents added before next request"
}

class DiagnosticMessage(BaseModel):
    timestep: int
    issue_type: int
    agent_list: Optional[List[int]] = None

class DiagnosticTool:
    """
    A user-side tool that examines a debug log checking for common implementation mistakes 
    and provides feedback.

    Parameters
    ----------
    debug_log_path:
        The full path to the debug log file to be loaded and analyzed.
    ego_indexes:
        A list of index IDs for ego vehicles that will be analyzed differently (e.g. it is expected
        that the state of an ego vehicle may be modified external to the API).
    """

    def __init__(
        self,
        debug_log_path: str,
        ego_indexes: List[int] = None
    ):
        self.debug_log_path = debug_log_path
        self.log_data = None
        with open(debug_log_path) as json_file:
            self.log_data = json.load(json_file)
        (
            self.req_groupings, 
            self.res_groupings,
            self.req_agent_details,
            self.init_agent_details
        ) = self._parse_log_data(log_data = self.log_data)
        self.num_timesteps = len(self.res_groupings["agent_states"])

        self.DIAGNOSTIC_ISSUE_LIBRARY = DIAGNOSTIC_ISSUE_LIBRARY

        self.ego_indexes = [] if ego_indexes is None else ego_indexes

    def _format_message(
        self,
        timestep: int,
        issue_type: str,
        agent_list: Optional[List[int]] = None
    ):
        return DiagnosticMessage(
            timestep=timestep,
            issue_type=issue_type,
            agent_list=agent_list
        )

    def full_diagnostic_test(self):
        """
        Main access point for this class. This function runs all available diagnostic tests and prints 
        human-readable feedback for the benefit of a user to fix issues in their integration.
        """

        diagnostic_message_codes = []

        diagnostic_message_codes.extend(self._check_drive_response_equivalence())
        diagnostic_message_codes.extend(self._check_number_of_agents())
        diagnostic_message_codes.extend(self._check_agent_details_realistic())
        diagnostic_message_codes.extend(self._check_agent_details_static())

        diagnostic_message_dict = {}
        for msg in diagnostic_message_codes:
            if msg.timestep in diagnostic_message_dict:
                diagnostic_message_dict[msg.timestep].append(msg)
            else:
                diagnostic_message_dict[msg.timestep] = [msg]

        print(f"===========================================================================================")
        print(f"Printing potential implementation issues that might cause performance degradation:")
        print(f"")

        for ts in range(self.num_timesteps):
            if ts not in diagnostic_message_dict:
                print(f"At timestep {ts}: No issues detected.")
            else:
                for msg in diagnostic_message_dict[ts]:
                    if msg.agent_list is not None:
                        print(f"At timestep {ts}: Issue code {msg.issue_type} applicable to agent IDs {msg.agent_list}.")
                    else:
                        print(f"At timestep {ts}: Issue code {msg.issue_type}.")

        print(f"")
        print(f"===========================================================================================")
        print(f"Diagnostic Issue Legend:")
        print(f"")
        for code, message in self.DIAGNOSTIC_ISSUE_LIBRARY.items():
            print(f"{code}:{message}")
        print(f"")
        print(f"===========================================================================================")
        print(f"")

        print(f"Finished diagnostic analysis.")

    def _check_agent_details_realistic(self):
        diagnostic_message_codes = []

        flagged_agents = []
        for ind, detes in enumerate(self.init_agent_details):
            if detes[3] == "car":
                if not ((3.0 < detes[0] < 7.0) and (1.0 < detes[1] < 3.0) and (detes[0]*0.05 < detes[2] < detes[0]*0.95)):
                    flagged_agents.append(ind)
            if detes[3] == "pedestrian":
                if not ((0.5 < detes[0] < 2.0) and (0.5 < detes[1] < 2.0)):
                    flagged_agents.append(ind)
        if len(flagged_agents) > 0:
            diagnostic_message_codes.append(
                self._format_message(
                    timestep = 0,
                    issue_type = 50,
                    agent_list = flagged_agents
                )
            )

        return diagnostic_message_codes

    def _check_agent_details_static(self):
        diagnostic_message_codes = []

        is_equal_agent_details = {"details_equal":[],"same_index":[]}
        all_agent_details = [self.init_agent_details] + self.req_agent_details

        for ts in range(len(all_agent_details)-1):
            details_equal, is_index_equal = self._check_states_equal(
                [prop[:-1] for prop in all_agent_details[ts]],
                [prop[:-1] for prop in all_agent_details[ts+1]]
            )
            is_equal_agent_details["details_equal"].append(details_equal)
            is_equal_agent_details["same_index"].append(is_index_equal)

        for ind, (ts_agent_exists, ts_index) in enumerate(zip(is_equal_agent_details["details_equal"],is_equal_agent_details["same_index"])):
            if not all(ts_agent_exists):
                diagnostic_message_codes.append(
                    self._format_message(
                        timestep = ind,
                        issue_type = 51,
                        agent_list = list(filter(lambda i: not ts_agent_exists[i], range(len(ts_agent_exists))))
                    )
                )
            elif not all(ts_index):
                diagnostic_message_codes.append(
                    self._format_message(
                        timestep = ind,
                        issue_type = 52,
                        agent_list = list(filter(lambda i: ts_agent_exists[i] and not ts_index[i], range(len(ts_index))))
                    )
                )

        return diagnostic_message_codes

    def _check_number_of_agents(self):
        diagnostic_message_codes = []

        for ind, (req_list, res_list) in enumerate(zip(self.req_groupings["agent_states"][1:],self.res_groupings["agent_states"][:-1])):
            if len(res_list) < len(req_list):
                diagnostic_message_codes.append(
                    self._format_message(
                        timestep = ind+1,
                        issue_type = 101
                    )
                )

        return diagnostic_message_codes

    def _check_drive_response_equivalence(self):
        diagnostic_message_codes = []

        is_equal_agent_states = {"agents_equal":[],"same_index":[]}
        is_equal_recurrent_states = {"agents_equal":[],"same_index":[]}

        for req_dict, res_dict in zip(self.req_groupings["agent_states"][1:],self.res_groupings["agent_states"][:-1]):
            states_equal, is_index_equal = self._check_states_equal(res_dict,req_dict)
            is_equal_agent_states["agents_equal"].append(states_equal)
            is_equal_agent_states["same_index"].append(is_index_equal)
        for req_dict, res_dict in zip(self.req_groupings["recurrent_states"][1:],self.res_groupings["recurrent_states"][:-1]):
            states_equal, is_index_equal = self._check_states_equal(res_dict,req_dict)
            is_equal_recurrent_states["agents_equal"].append(states_equal)
            is_equal_recurrent_states["same_index"].append(is_index_equal)

        for ind, (ts_agent_exists, ts_index) in enumerate(zip(is_equal_agent_states["agents_equal"],is_equal_agent_states["same_index"])):
            if not all(ts_agent_exists):
                diagnostic_message_codes.append(
                    self._format_message(
                        timestep = ind+1,
                        issue_type = 11,
                        agent_list = list(filter(lambda i: not ts_agent_exists[i], range(len(ts_agent_exists))))
                    )
                )
            elif not all(ts_index):
                diagnostic_message_codes.append(
                    self._format_message(
                        timestep = ind+1,
                        issue_type = 10,
                        agent_list = list(filter(lambda i: ts_agent_exists[i] and not ts_index[i], range(len(ts_index))))
                    )
                )

        for ind, (ts_agent_exists, ts_index) in enumerate(zip(is_equal_recurrent_states["agents_equal"],is_equal_recurrent_states["same_index"])):
            if not all(ts_agent_exists):
                diagnostic_message_codes.append(
                    self._format_message(
                        timestep = ind+1,
                        issue_type = 21,
                        agent_list = list(filter(lambda i: not ts_agent_exists[i], range(len(ts_agent_exists))))
                    )
                )
            elif not all(ts_index):
                diagnostic_message_codes.append(
                    self._format_message(
                        timestep = ind+1,
                        issue_type = 20,
                        agent_list = list(filter(lambda i: ts_agent_exists[i] and not ts_index[i], range(len(ts_index))))
                    )
                )

        return diagnostic_message_codes

    def _get_agent_details(
        self,
        agent_dict: Dict
    ):
        ts_agent_details = []

        agent_details = []
        #Covers cases where agent attributes is either None or an empty list
        if "agent_properties" in agent_dict:
            for detes in agent_dict["agent_properties"]:
                ts_agent_details.append([
                    detes["length"],
                    detes["width"],
                    detes["rear_axis_offset"],
                    detes["agent_type"],
                    detes["waypoint"]
                ])
        elif "agent_attributes" in agent_dict:
            ts_agent_details = agent_dict["agent_attributes"]            

        return ts_agent_details

    def _parse_log_data(
        self,
        log_data: Dict
    ):
        STATE_DECIMAL = 2
        RECURR_DECIMAL = 6

        req_agent_state_dict = {"agent_states":[],"recurrent_states":[]}
        res_agent_state_dict = {"agent_states":[],"recurrent_states":[]}
        req_agent_details = []
        init_agent_details = []

        if "large_drive_responses" in log_data:
            drive_req_data = log_data["large_drive_requests"]
            drive_res_data = log_data["large_drive_responses"]
        else:
            drive_req_data = log_data["drive_requests"]
            drive_res_data = log_data["drive_responses"]

        for req_json, res_json in zip(drive_req_data,drive_res_data):
            res_data = json.loads(res_json)
            req_data = json.loads(req_json)

            for (state_dict, data) in zip([req_agent_state_dict,res_agent_state_dict],[req_data,res_data]):
                state_dict["agent_states"].append([[round(x,STATE_DECIMAL) for x in st] for st in data["agent_states"]])
                recurr_state = None
                if data["recurrent_states"] is not None:
                    recurr_state = [[round(x,RECURR_DECIMAL) for x in st] for st in data["recurrent_states"]]
                state_dict["recurrent_states"].append(recurr_state)

        for req_json in drive_req_data:
            req_data = json.loads(req_json)
            req_agent_details.append(self._get_agent_details(req_data))
        
        if "large_initialize_responses" in log_data:
            init_res_data = log_data["large_initialize_responses"]
            res_data = json.loads(init_res_data[-1])
            init_agent_details = self._get_agent_details(res_data)
        elif "initialize_responses" in log_data:
            init_res_data = log_data["initialize_responses"]
            res_data = json.loads(init_res_data[-1])
            init_agent_details = self._get_agent_details(res_data)
        else:
            init_agent_details = req_agent_details.pop(0)

        return req_agent_state_dict, res_agent_state_dict, req_agent_details, init_agent_details

    def _check_states_equal(
        self,
        res_states: List[Union[float,str,List[float]]],
        req_states: Optional[List[Union[float,str,List[float]]]] = None
    ):
        num_res_agents = len(res_states)
        states_equal = [False]*num_res_agents
        is_same_index = [False]*num_res_agents

        if res_states is not None and req_states is not None:
            for i, sa_0 in enumerate(res_states):
                is_state_equal = False
                is_index_equal = False
                for j, sa_1 in enumerate(req_states):
                    is_state_equal = sa_0 == sa_1
                    if is_state_equal: 
                        is_index_equal = i == j
                        break
                
                # Some behaviour is expected with ego agents controlled externally to the API
                states_equal[i] = is_state_equal or i in self.ego_indexes
                is_same_index[i] = is_index_equal or i in self.ego_indexes

        return states_equal, is_same_index

if __name__ == '__main__':
    argparser = argparse.ArgumentParser(description=__doc__)
    argparser.add_argument(
        '--debug_log_path',
        type=str,
        help=f"Full path to an IAI debug log to be analyzed.",
        default='None'
    )
    argparser.add_argument(
        '--ego_agent_list',
        type=int,
        nargs='+',
        help=f"IDs of ego agents.",
        default=None
    )
    args = argparser.parse_args()

    diagnostic_tool = DiagnosticTool(
        debug_log_path=args.debug_log_path,
        ego_indexes=args.ego_agent_list
    )
    diagnostic_tool.full_diagnostic_test()


import invertedai as iai
"""
Inspired by OpenAI python-API error handling
"""


class InvertedAIError(Exception):
    """
    Base exception class for Inverted AI Python-API error handling.
    """

    def __init__(
        self,
        message=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
    ):
        super(InvertedAIError, self).__init__(message)

        if http_body and hasattr(http_body, "decode"):
            try:
                http_body = http_body.decode("utf-8")
            except BaseException:
                http_body = (
                    "<Could not decode body as utf-8. "
                    "Please report to info@inverted.ai>"
                )

        self._message = message
        self.http_body = http_body
        self.http_status = http_status
        self.json_body = json_body
        self.headers = headers or {}
        self.code = code

    def __str__(self):
        msg = self._message or "<empty message>"
        return msg

    @property
    def user_message(self):
        return self._message

    def __repr__(self):
        return "%s(message=%r, http_status=%r, request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.http_status,
        )


class APIError(InvertedAIError):
    """
    Errors which are caused by incorrect API call.
    """
    pass


class TryAgain(InvertedAIError):
    """
    Errors that may be resolved upon resending the request.
    """
    pass


class InvalidAPIKeyError(InvertedAIError):
    """
    Missing API Key or invalid API Key type.
    """
    pass


class APIConnectionError(InvertedAIError):
    """
    API-gateway error due to invalid API Key.
    """

    def __init__(
        self,
        message,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
        should_retry=False,
    ):
        super(APIConnectionError, self).__init__(
            message, http_body, http_status, json_body, headers, code
        )
        self.should_retry = should_retry


class InvalidRequestError(InvertedAIError):
    """
    Bad Request error (for example, malformed request syntax,
    invalid request message framing, or deceptive request routing).
    """

    def __init__(
        self,
        message,
        param,
        code=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
    ):
        super(InvalidRequestError, self).__init__(
            message, http_body, http_status, json_body, headers, code
        )
        self.param = param

    def __repr__(self):
        return "%s(message=%r, param=%r, code=%r, http_status=%r, " "request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.param,
            self.code,
            self.http_status,
        )

    def __reduce__(self):
        return type(self), (
            self._message,
            self.param,
            self.code,
            self.http_body,
            self.http_status,
            self.json_body,
            self.headers,
        )


class ResourceNotFoundError(InvertedAIError):
    pass


class AuthenticationError(InvertedAIError):
    """
    API-gateway error due to invalid API Key.
    """
    pass


class PermissionError(InvertedAIError):
    pass


class RateLimitError(InvertedAIError):
    pass


class ServiceUnavailableError(InvertedAIError):
    pass


class InvalidAPIType(InvertedAIError):
    pass

class RequestTooLarge(InvertedAIError):
    pass

class RequestTimeoutError(InvertedAIError):
    pass

class ServerTimeoutError(InvertedAIError):
    pass


class SignatureVerificationError(InvertedAIError):
    def __init__(self, message, sig_header, http_body=None):
        super(SignatureVerificationError, self).__init__(message, http_body)
        self.sig_header = sig_header

    def __reduce__(self):
        return type(self), (
            self._message,
            self.sig_header,
            self.http_body,
        )


class InvalidInput(InvertedAIError):
    """
    Invalid Python API input.
    """
    pass


class InvalidInputType(InvalidInput):
    """
    Invalid input type for Python API.
    """
    pass



from typing import List, Optional, Union
from copy import deepcopy

import invertedai as iai
from invertedai.common import (
    AgentProperties,
    AgentState, 
    RecurrentState,
    TrafficLightStatesDict
)
from invertedai.large.drive import large_drive
from invertedai.large.initialize import large_initialize, get_regions_default
from invertedai.api.drive import DriveResponse
from invertedai.api.initialize import InitializeResponse

class BasicCosimulation:
    """
    Stateful wrapper around the Inverted AI API to simplify co-simulation.
    All keyword arguments to :func:`large_initialize` can be passed to the constructor 
    here and a sufficient combination of them must be passed as required by :func:`large_initialize`.
    This wrapper caches static agent properties and propagates the recurrent state,
    so that only states of conditional agents need to be exchanged with it to
    perform co-simulation. Typically, each time step requires a single call to :func:`self.step`.

    This wrapper only supports a minimal co-simulation functionality.
    For more advanced use cases, call :func:`large_initialize` and :func:`large_drive` directly.

    location: 
        Location name as expected by :func:`large_initialize` and :func:`large_drive`.
    conditional_agent_properties: 
        Agent properties for all conditional agents that must be given to the API so the NPC agents are aware of them. 
        Please refer to the documentation for :func:`large_initialize` for more information on how to format this parameter (treating the
        conditional agents as "predefined agents"). Furthermore, any predefined agents for which the user
        wishes the IAI API to control must be defined at the end of this list.
    conditional_agent_agent_states: 
        Agent states for all conditional agents that must be given to the API so the NPC agents are aware of their states. 
        Please refer to the documentation for :func:`large_initialize` for more information on how to format this parameter (treating the
        ego agents as "predefined agents"). Furthermore, any predefined agents for which the user
        wishes the IAI API to control must be defined at the end of this list.
    num_non_ego_conditional_agents: 
        The ego agents are the subset of the conditional agents that are NOT controlled by the Inverted AI API. This
        parameter allows some of the conditional agents with predefined states and properties to nonetheless be 
        controlled by the Inverted AI API. The non-ego conditional agents must be placed at the end of the conditional
        agents list and the ego agents must be placed at the beginning of the conditional agents list.
    """

    def __init__(
        self,
        location: str,
        conditional_agent_properties: Optional[List[AgentProperties]] = None,
        conditional_agent_agent_states: Optional[List[AgentState]] = None,
        num_non_ego_conditional_agents: Optional[int] = 0,
        **kwargs # sufficient arguments to initialize must also be included
    ):
        self._conditional_agent_properties = conditional_agent_properties
        self._conditional_agent_agent_states = conditional_agent_agent_states

        self._location = location
        self._response = large_initialize(
            location=self._location,
            agent_properties=self._conditional_agent_properties,
            agent_states=self._conditional_agent_agent_states,
            **kwargs,
        )
        self.init_response = deepcopy(self._response)
        self._light_state = self.init_response.traffic_lights_states
        self._light_recurrent_state = self.init_response.light_recurrent_states

        self._total_agent_count = len(self.init_response.agent_properties)  # initialize may produce different agent count
        self._conditional_agent_count = len(self._conditional_agent_agent_states) - num_non_ego_conditional_agents
        self._npc_agent_count = self._total_agent_count - self._conditional_agent_count
        assert self._conditional_agent_count >= 0, "Invalid number of ego and conditional agents."
        
        self._agent_properties = self.init_response.agent_properties
        self._agent_states = self.init_response.agent_states
        self._recurrent_states = self.init_response.recurrent_states
        
    @property
    def location(self) -> str:
        """
        Location name as recognized by Inverted AI API.
        """
        return self._location

    @property
    def agent_count(self) -> int:
        """
        The total number of agents, both ego and NPCs.
        """
        return self._total_agent_count

    @property
    def agent_states(self) -> List[AgentState]:
        """
        The predicted states for all agents, including ego.
        """
        return self._agent_states

    @property
    def agent_properties(self) -> List[AgentProperties]:
        """
        The properties (length, width, rear_axis_offset, max_speed) for all agents, including ego.
        """
        return self._agent_properties

    @property
    def ego_states(self) -> List[AgentState]:
        """
        Returns the predicted states of ego agents in order.
        The NPC agents are excluded.
        """
        return self._agent_states[:self._conditional_agent_count]

    @property
    def ego_properties(self) -> List[AgentProperties]:
        """
        Returns the properties of ego agents in order.
        The NPC agents are excluded.
        """
        return self._agent_properties[:self._conditional_agent_count]

    @property
    def npc_states(self) -> List[AgentState]:
        """
        Returns the predicted states of NPCs (non-ego agents) in order.
        The predictions for ego agents are excluded.
        """
        return self._agent_states[self._conditional_agent_count:]

    @property
    def npc_properties(self) -> List[AgentProperties]:
        """
        Returns the properties of NPCs (non-ego agents) in order.
        The ego agents are excluded.
        """
        return self._agent_properties[self._conditional_agent_count:]

    @property
    def npc_recurrent_states(self) -> List[RecurrentState]:
        """
        Returns the recurrent states of NPCs (non-ego agents) in order.
        The ego agents are excluded.
        """
        return self._recurrent_states[self._conditional_agent_count:]

    @property
    def light_states(self) -> Optional[TrafficLightStatesDict]:
        """
        Returns the traffic light states if any exists on the map.
        """
        return self._light_state

    @property
    def response(self) -> Union[DriveResponse,InitializeResponse]:
        """
        Get the current response data object containing all information received from the API.
        """
        return self._response

    def step(
        self, 
        current_conditional_agent_states: List[AgentState],
        **kwargs
    ) -> None:
        """
        Calls :func:`large_drive` to advance the simulation by one time step.
        Current states of ego agents need to be provided to synchronize with
        your local simulator.
        This function assumes the conditional agents are placed at the beginning
        of given agent states list.
        All remaining keyword arguments to :func:`large_drive` can be passed 
        to this function here to receive the desired information from the API.

        current_conditional_agent_states:  
            States of ego agents before the step which must match the number of 
            given ego, conditional agents during initialization.
        """
        self._update_conditional_states(current_conditional_agent_states)
        
        self._response = large_drive(
            location=self.location,
            agent_properties=self._agent_properties,
            agent_states=self._agent_states,
            recurrent_states=self._recurrent_states,
            light_recurrent_states=self._light_recurrent_state,
            **kwargs
        )
        self._agent_states = self._response.agent_states
        self._recurrent_states = self._response.recurrent_states
        self._light_state = self._response.traffic_lights_states
        self._light_recurrent_state = self._response.light_recurrent_states

    def _update_conditional_states(self, conditional_agent_states):
        assert len(conditional_agent_states) == self._conditional_agent_count, "Given number of agents in this step must match the number of ego agents in the co-simulation."

        self._agent_states[:self._conditional_agent_count] = conditional_agent_states
